#include <stdio.h>
#include <ctype.h>

#include "mbee77.h"
#include "rawmem.h"
#include "humutil.h"

// Defines for the menu lists
#define MAIN_MENU           0

// Defines for the switch statements for Main Menu
#define MAIN_SELECT_DEVICE  0
#define MAIN_QUIT           1

// Number of devices to list
#define MAXDEVICES          10

// Family codes
#define DS1921_FC           0x21
#define DS1922_23_FC        0x41

// Data log size
#define DS1921_SIZE         2048
#define DS1922_23_SIZE      8192


int main(int argc, char **argv)
{
    int done = FALSE;
    int portnum = 0;
    uchar AllSN[MAXDEVICES][8];
    int NumDevices;
    int owd;
    char msg[256];
    configLog calData;
    int ds1923;

    uchar dataLogAry[DS1922_23_SIZE];
    
    FILE* datafp;
    int resolution;

    if (argc != 2)
    {
        sprintf(msg,"1-Wire Net name required on command line!\n"
                " (example: \"COM1\" (Win32 DS2480),\"/dev/cua0\" "
                "(Linux DS2480),\"{1,5}\" (Win32 TMEX)\n"
                "\"DS2490-0\" (WinUSB), \"ds2490-1\" (libusb)");

        printf("%s\n",msg);
        return 0;
    }

    printf("\niButton Data Log Dump for DS1921G, DS1922L/T, DS1923\n");

   if((portnum = owAcquireEx(argv[1])) < 0)
   {
      OWERROR_DUMP(stdout);
      return 0;
   }
   else
   {
      // loop to do menu
      do
      {
          // Main menu
         switch (menuSelect(MAIN_MENU,&AllSN[0][0]))
         {
            case MAIN_SELECT_DEVICE :
                // find all parts
                // loop to find all of the devices up to MAXDEVICES
                NumDevices = 0;

                do
                {
                    // perform the search
                    if (!owNext(portnum,TRUE, FALSE))
                        break;

                    owSerialNum(portnum,AllSN[NumDevices], TRUE);
                    NumDevices++;
                }
                while (NumDevices < (MAXDEVICES - 1));
            
                // select a device
                owd = selectDevice(NumDevices,&AllSN[0]);

                // high or low resolution
                resolution = getResolution(AllSN[owd]);

                // read calibration data for ds1922, ds1923
                if(AllSN[owd][0] == DS1922_23_FC)
                    readCalibration(portnum, AllSN[owd], &calData);

                // for ds1923 devices, ask if humidity was logged
                ds1923 = getHumidity(AllSN[owd], calData);

                // output data files
                datafp = fopen("memlog_temperture.txt", "a");

                fprintf(datafp, "iButton Data Log Dump\n");
                fprintf(datafp, "-----------------------------\n");

                printf("Reading iButton Data, please wait...\n");

                // read the raw memory from iButton
                readDataLog(portnum, AllSN[owd], &dataLogAry);


                // interpret data based on temperature or temperature/humidity
                if(ds1923 == 0)
                {
                    dataLog(dataLogAry, resolution, AllSN[owd], calData, 
                            datafp);
                }
                else
                {
                    // Setup data structure for humidity cal calculations
                    if(resolution == 1)
                        calData.adDeviceBits = 8;
                    else if(resolution == 2)
                        calData.adDeviceBits = 10; // OneWireViewer uses 10

                    dataHumaLog(dataLogAry, resolution, AllSN[owd], calData, 
                            datafp);
                }

                fclose(datafp);

                printf("Finished\n");

            case MAIN_QUIT:
                done = TRUE;
                break;
          }
      }
      while(!done);
   }
}

/**
 * Reads data log memory of iButton into buffer.
 *
 * portnum  1-Wire adapter
 * SNum     iButton serial number of interest
 * buffer   empty uchar array, should have a size of 8192
 *
 * @return read status
 */
int readDataLog(int portnum, uchar SNum[8], uchar *buffer)
{
    int i;

    if(SNum[0] == DS1921_FC) // DS1921
    {
        for(i=0; i<64; i++) // 64 pages for data log
        {
            if(readPageNVCRC(5, // data log bank
                    portnum,
                    SNum,
                    i, // page number
                    FALSE, // no read cont
                    &buffer[i*32]) == FALSE)
                return FALSE;
        }
    } 
    else if(SNum[0] == DS1922_23_FC) // DS1922, DS1923
    {
        for(i=0; i<256; i++)
        {
            // 128 data-log memory offset
            if(readPageCRCEE77(0, portnum, SNum, i+128, &buffer[i*32]) 
                    == FALSE) 
                return FALSE;
        }
    }

    return TRUE;
}

/**
 * Read and calculation calibration coefficients for DS1922 and DS1923.
 * Temperature and humidity correction factors.
 *
 * portnum  1-Wire adapater
 * SNum     iButton serial number of interest
 * calData  struct pointer to store the calibration coefficients
 *
 * @return calculation and read successful
 */
int readCalibration(int portnum, uchar SNum[8], configLog* calData)
{
    uchar temp_buff[96];

    double ref1 = 0.0, ref2 = 0.0, ref3 = 0.0;
    double read1 = 0.0, read2 = 0.0, read3 = 0.0;
    double error1 = 0.0, error2 = 0.0, error3 = 0.0;
    double ref1sq, ref2sq, ref3sq;

    if(readPageCRCEE77(0,portnum,SNum,16,&temp_buff[0]) == FALSE)
        return FALSE;
    if(readPageCRCEE77(0,portnum,SNum,17,&temp_buff[32]) == FALSE)
        return FALSE; 
    if(readPageCRCEE77(0,portnum,SNum,18,&temp_buff[64]) == FALSE)
        return FALSE;
    
    calData->configByte = temp_buff[32+6];

    if(calData->configByte == 0x20) // ds1923
    {
        // setup constants for function decodeHumidity
        calData->adDeviceBits = 10;
        calData->adReferenceVoltage = 5.02;
        calData->lowTemp = -40; // the spec is -20, -40 is for calculation
        calData->highTemp = 85;

        ref1 = decodeHumidity(&temp_buff[72],2,TRUE,*calData);
        read1 = decodeHumidity(&temp_buff[74],2,TRUE,*calData);
        error1 = read1 - ref1;
        ref2 = decodeHumidity(&temp_buff[76],2,TRUE,*calData);
        read2 = decodeHumidity(&temp_buff[78],2,TRUE,*calData);
        error2 = read2 - ref2;
        ref3 = decodeHumidity(&temp_buff[80],2,TRUE,*calData);
        read3 = decodeHumidity(&temp_buff[82],2,TRUE,*calData);
        error3 = read3 - ref3;

        ref1sq = ref1*ref1;
        ref2sq = ref2*ref2;
        ref3sq = ref3*ref3;

        calData->humCoeffB =
            ( (ref2sq-ref1sq)*(error3-error1) + ref3sq*(error1-error2)
              + ref1sq*(error2-error1) ) /
            ( (ref2sq-ref1sq)*(ref3-ref1) + (ref3sq-ref1sq)*(ref1-ref2) );
        calData->humCoeffA =
            ( error2 - error1 + calData->humCoeffB*(ref1-ref2) ) /
            ( ref2sq - ref1sq );
        calData->humCoeffC =
            error1 - calData->humCoeffA*ref1sq - calData->humCoeffB*ref1;
    }
    else if(calData->configByte == 0x40) // ds1922l
    {
        calData->lowTemp = -40;
        calData->highTemp = 85;
        calData->hasHumidity = FALSE;
    }
    else if(calData->configByte == 0x60) // ds1922t
    {
        calData->lowTemp = 0;
        calData->highTemp = 125;
        calData->hasHumidity = FALSE;
    }
    else if(calData->configByte == 0x80) // ds1922e
    {
        // Not implemented
    }

    // temperature calibration - for ds1922l/ds1922t/ds1923 devices
    ref2 = decodeTemperature(&temp_buff[64],2,TRUE,*calData);
    read2 = decodeTemperature(&temp_buff[66],2,TRUE,*calData);
    error2 = read2 - ref2;
    ref3 = decodeTemperature(&temp_buff[68],2,TRUE,*calData); read3 = decodeTemperature(&temp_buff[70],2,TRUE,*calData);
    error3 = read3 - ref3;
    if(calData->configByte == 0x60) // ref value constant from datasheet
        ref1 = 90.0;
    else
        ref1 = 60.0;
    error1 = error2;
    read1 = ref1 + error1;

    ref1sq = ref1*ref1;
    ref2sq = ref2*ref2;
    ref3sq = ref3*ref3;

    calData->tempCoeffB =
        ( (ref2sq-ref1sq)*(error3-error1) + ref3sq*(error1-error2)
          + ref1sq*(error2-error1) ) /
        ( (ref2sq-ref1sq)*(ref3-ref1) + (ref3sq-ref1sq)*(ref1-ref2) );
    calData->tempCoeffA =
        ( error2 - error1 + calData->tempCoeffB*(ref1-ref2) ) /
        ( ref2sq - ref1sq );
    calData->tempCoeffC =
        error1 - calData->tempCoeffA*ref1sq - calData->tempCoeffB*ref1;  

    return TRUE;
}

/**
 * Reads and saves temperature data log for DS1921, DS1922 to file.
 *
 * rawData  Array which stores the iButton data log
 * resBytes Number of bytes for a temperature point
 * SNum     iButton serial number of interest
 * calData  struct containing the calibration coefficients
 * log      FILE* for storing data
 *
 * return   read status
 */
int dataLog(uchar* rawData, int resBytes, uchar SNum[8], configLog calData, 
        FILE* log)
{
    int i;
    double temperature;
    double temperatureCal;

    if(SNum[0] == DS1921_FC) // ds1921 - single resolution
    {
        for(i=0; i<DS1921_SIZE; i++)
        {
            temperature = (rawData[i] & 0x0FF)/2.0 -40;
            fprintf(log, "%.2f\n", temperature);
        }
    }
    else if(SNum[0] == DS1922_23_FC) // ds1922, ds1923 - multiple resolutions
    {
        for(i=0; i<DS1922_23_SIZE/resBytes; i++)
        {
            if(resBytes == 1)
                temperature = (rawData[i*resBytes] & 0x0FF)/2.0 - 41;
            else if(resBytes == 2)
                temperature = (rawData[i*resBytes] & 0x0FF)/2.0 - 41 
                    + (rawData[i*resBytes + 1] & 0x0FF)/512.0; 
            else
                return FALSE; // non-existent resolution

            temperatureCal = temperature 
                - (calData.tempCoeffA*temperature*temperature 
                        + calData.tempCoeffB*temperature + calData.tempCoeffC);

            fprintf(log, "%.3f\n", temperatureCal);
        }
    }
    else
        return FALSE; // non iButton logger family code

    return TRUE;
}

/**
 * Reads and saves temperature and humidity data log for DS1923to file.
 *
 * rawData  Array which stores the iButton data log
 * resBytes Number of bytes for a temperature point
 * SNum     iButton serial number of interest
 * calData  struct containing the calibration coefficients
 * log      FILE* for storing data
 *
 * return   read status
 */
int dataHumaLog(uchar* rawData, int resBytes, uchar SNum[8], configLog calData,
        FILE* log)
{
    int i;
    double temperature, humidity;
    double temperatureCal, humidityCal;

    for(i=0; i<(DS1922_23_SIZE/resBytes)/2; i++) // humidity and temperature
    {
        if(resBytes == 1)
            temperature = (rawData[i*resBytes] & 0x0FF)/2.0 - 41;
        else if(resBytes == 2)
            temperature = (rawData[i*resBytes] & 0x0FF)/2.0 - 41 
                + (rawData[i*resBytes + 1] & 0x0FF)/512.0; 
        else
            return FALSE; // non-existent resolution

        temperatureCal = temperature 
            - (calData.tempCoeffA*temperature*temperature 
                    + calData.tempCoeffB*temperature + calData.tempCoeffC);

        // Humidity data is located in the second half of rawData[]
        humidity = decodeHumidity(&rawData[i*resBytes + DS1922_23_SIZE/2], 
                resBytes, TRUE, calData);

        humidityCal = humidity - (calData.humCoeffA*humidity*humidity
                + calData.humCoeffB*humidity + calData.humCoeffC);

        fprintf(log, "%.3f\t%.3f\n", temperatureCal, humidityCal);
    }

    return TRUE;
}


//--------
//-------- Menu Methods
//--------

/**
 * Display menu and ask for a selection.
 *
 * menu  the menu selections.
 *
 * @return numberic value entered from the console.
 */
int menuSelect(int menu, uchar *SNum)
{
    int length = 0;

    switch(menu)
    {
        case MAIN_MENU:
            printf("\n");
            printf("MainMenu 1-Wire Memory Demo\n");
            printf("(0) Select Device\n");
            printf("(1) Quit\n");
            length = 1;
            break;

    }

    printf("\nPlease enter value: ");

    return getNumber(0, length);
}

/**
 * Retrieve user input from the console.
 *
 * min  minimum number to accept
 * max  maximum number to accept
 *
 * @return numeric value entered from the console.
 */
int getNumber (int min, int max)
{
    int value = min,cnt;
    int done = FALSE;
    do
    {
        cnt = scanf("%d",&value);
        if(cnt>0 && (value>max || value<min))
        {
            printf("Value (%d) is outside of the limits (%d,%d)\n",value,min,max);
            printf("Try again:\n");
        }
        else
            done = TRUE;

    }
    while(!done);

    return value;

}

int getResolution(uchar SNum[8])
{
    if(SNum[0] == DS1921_FC)
        return 1;
    else if(SNum[0] == DS1922_23_FC)
    {
        printf("Resolution setting (1 for 0.5C, 2 for 0.0625C): ");
        return getNumber(1,2);
    }

    return 1;
}

int getHumidity(uchar SNum[8], configLog config)
{
    if((SNum[0] == DS1922_23_FC) && (config.configByte == 0x20))
    {
        printf("Did you log humidity data? (1 for yes, 0 for no): ");
        return getNumber(0,1);
    }

    return 0;
}

//--------
//-------- Menu methods
//--------

/**
 * Create a menu from the provided OneWireContainer
 * Vector and allow the user to select a device.
 *
 * numDevices  the number of devices on the 1-Wire.
 * AllDevices  holds all the Serial Numbers on the
 *                      1-Wire.
 *
 * @return index of the chosen device.
 */
SMALLINT selectDevice(int numDevices, uchar AllDevices[][8])
{

   int i,j;

   printf("Device Selection\n");

   for (i=0; i<numDevices; i++)
   {
      printf("(%d) ",i);
      for(j=7;j>=0;j--)
         printf("%02X ",AllDevices[i][j]);
      printf("\n");
   }

   return getNumber(0,numDevices-1);
}
